__all__ = [
    "__title__", "__summary__", "__uri__", "__version__", "__author__",
    "__email__", "__license__", "__copyright__",
]

__title__ = "spmat"
__summary__ = "spmat: A collection of tools for special matrices"
__uri__ = "https://github.com/zhengp0/spmat"

__version__ = "0.0.7"

__author__ = "Peng Zheng"
__email__ = "zhengp@uw.edu"

__license__ = "BSD 2-Clause License"
__copyright__ = f"Copyright 2021 {__author__}"
